(******************************************************************************)
(* generator                                                                  *)
(* (C) 2010..2021 Steffen Solyga <solyga@gmx.de>                              *)
(******************************************************************************)
IMPLEMENTATION MODULE Gen;

IMPORT FIO;
IMPORT Comp;

CONST
  debug = TRUE;
  out ::= FIO.StandardOutput;
  err ::= FIO.ErrorOutput;
  dbg ::= FIO.ErrorOutput;

  ocMax = 3;   (* 3 operands are enough for l04 (no relations/conditions) *)

TYPE
  TStr = ARRAY [ 0 .. 2 ] OF CHAR;

VAR
  oc: CARDINAL;  (* operand count *)
  vc: CARDINAL;  (* variable count *)
  av: CARDINAL;  (* accumulator value (0..Comp.cnt-1: var idx, Comp.cnt: val *)
  oprs: ARRAY TOperation OF TStr;


PROCEDURE Dbg0();
BEGIN
  FIO.WrStr( dbg, 'Gen.Debug: vars operands' );
  FIO.WrLn( dbg );
END Dbg0;

PROCEDURE DbgS( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( dbg, 'Gen.Debug:  ' );
  FIO.WrCard( dbg, vc, 2 );
  FIO.WrStr( dbg, '    ' );
  FIO.WrCard( dbg, oc, 1 );
  FIO.WrStr( dbg, ':  ' );
  FIO.WrStr( dbg, s );
END DbgS;

PROCEDURE DbgX();
BEGIN
  FIO.WrLn( dbg );
END DbgX;

PROCEDURE DbgV( v: TValue );
BEGIN
  FIO.WrStr( dbg, '  ' );
  FIO.WrInt( dbg, v, 0 );
  FIO.WrLn( dbg );
END DbgV;

PROCEDURE DbgO( o: TOffset );
BEGIN
  FIO.WrStr( dbg, '   ' );
  FIO.WrCard( dbg, o, 1 );
  FIO.WrLn( dbg );
END DbgO;


PROCEDURE Err( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( err, 'Gen: ' );
  FIO.WrStr( err, s );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END Err;

PROCEDURE Value( v: TValue ): Comp.TValue;
BEGIN
  RETURN v;
END Value;

PROCEDURE Index( o: TOffset ): Comp.TIndex;
BEGIN
  INC( o, ocMax );  (* ocMax vars reserved for calc stack *)
  IF o >= Comp.cnt THEN Err( 'No such variable' ) END;
  RETURN o;
END Index;

PROCEDURE Ldv( v: Comp.TValue );
BEGIN
  Comp.Ldv( v );
  av := Comp.cnt;
END Ldv;

PROCEDURE Ldi( i: Comp.TIndex );
BEGIN
  IF av # i THEN Comp.Ldi( i ) END;
  av := i;
END Ldi;

PROCEDURE Sti( i: Comp.TIndex );
BEGIN
  IF av # i THEN Comp.Sti( i ) END;
  av := i;
END Sti;


PROCEDURE Val( v: TValue );
BEGIN
  IF debug THEN DbgS( 'Val' ); DbgV( v ) END;
  IF oc >= ocMax THEN Err( 'Too many operands' ) END;
  Ldv( Value(v) ); Sti( oc ); INC( oc );
END Val;

PROCEDURE Lod( o: TOffset );
BEGIN
  IF debug THEN DbgS( 'Lod' ); DbgO( o ) END;
  IF oc >= ocMax THEN Err( 'Too many operands' ) END;
  Ldi( Index(o) ); Sti( oc ); INC( oc );
END Lod;

PROCEDURE Dec( x: CARDINAL );
BEGIN
  IF oc < x THEN Err( 'Too little operands' ) END;
  DEC( oc, x );
END Dec;

PROCEDURE Sto( o: TOffset );
BEGIN
  IF debug THEN DbgS( 'Sto' ); DbgO( o ) END;
  Dec( 1 ); Ldi( oc ); Sti( Index(o) );
END Sto;

PROCEDURE Calc( o: Comp.TOperation );
BEGIN
  Dec( 2 ); Ldi( oc );
  Comp.Opi( o, oc + 1 ); av := Comp.cnt;
  Sti( oc ); INC( oc );
END Calc;

PROCEDURE Opr( o: TOperation );
BEGIN
  IF debug THEN DbgS( oprs[o] ); DbgX() END;
  CASE o OF
  | oNeg: Dec(1); Ldi(oc); Comp.Neg(); av := Comp.cnt; Sti(oc); INC(oc);
  | oWrt: Dec(1); Ldi(oc); Comp.Out();
  | oAdd: Calc( Comp.oAdd );
  | oSub: Calc( Comp.oSub ); 
  | oMul: Calc( Comp.oMul );
  | oDiv: Calc( Comp.oDiv );
  | oNew: INC( vc ); IF vc > Comp.cnt THEN Err( 'Too many variables' ) END;
  ELSE
    Err( 'Bug: Invalid operation' );
  END;
END Opr;

BEGIN
  oc := 0;
  vc := ocMax;      (* ocMax vars reserved for target stack *)
  av := Comp.cnt;
  oprs[ oNeg ] := 'Neg';
  oprs[ oWrt ] := 'Wrt';
  oprs[ oAdd ] := 'Add';
  oprs[ oSub ] := 'Sub';
  oprs[ oMul ] := 'Mul';
  oprs[ oDiv ] := 'Div';
  oprs[ oNew ] := 'New';
  IF debug THEN Dbg0() END;
END Gen.
